#    powernapd plugin - Monitors a UDP socket for data
#
#    Copyright (C) 2011 Canonical Ltd.
#
#    Authors: Dustin Kirkland <kirkland@canonical.com>
#             Adam Sutton <dev@adamsutton.me.uk>
#             Andres Rodriguez <andreserl@canonical.com>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import threading, time
from logging import error, debug, info, warn

# Monitor plugin
#   listen for data on a UDP socket (typically WOL packets)
class UDPMonitor (threading.Thread):

    # Initialise
    def __init__ ( self, config ):
        threading.Thread.__init__(self)
        self._type = config['monitor']
        self._name = config['name']
        self._port = config['port']
        self._running = False
        self._data_received = False
        self._absent_seconds = 0

    # Start thread
    def start ( self ):
      self._running = True
      threading.Thread.start(self)

    # Stop thread
    def stop ( self ): self._running = False

    # Open port and wait for data (any data will trigger the monitor)
    def run ( self ):
        import socket

        # Create socket
        sock   = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        listen = False

        while self._running:
            if not listen:
                try:
                    debug('%s - configure socket' % self)
                    sock.bind(('', self._port))
                    sock.settimeout(1.0)
                    listen = True
                except Exception, e:
                    error('%s - failed to config socket [e=%s]' % (self, str(e)))
                    time.sleep(1.0)
            else:
                try:
                    # Wait for data
                    sock.recvfrom(1024)
                    self._data_received = True
                    debug('%s - data packet received' % self)
                    self.reset()
                except: pass # timeout

    def active(self):
        if self._data_received:
            self._data_received = False
            return True
        return False

# ###########################################################################
# Editor directives
# ###########################################################################

# vim:sts=4:ts=4:sw=4:et



#    powernapd plugin - Abastract monitor class
#
#    Copyright (C) 2009 Canonical Ltd.
#
#    Authors: Dustin Kirkland <kirkland@canonical.com>
#             Adam Sutton <dev@adamsutton.me.uk>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

from logging import error, debug, info, warn
import time

# Abstract monitor
#   Note: it is not required that you subclass this, merely that you
#         provide a matching API (all functions are optional) 
class Monitor ( object ):

    # Initialise the object
    def __init__ ( self, config ):
        self._name     = self.__repr__()
        self._activity = 0
        self._grace    = 0
        self._period   = 60.0
        if ( config.has_key('name')   ): self._name   = config['name']
        if ( config.has_key('grace')  ): self._grace  = config['grace']
        if ( config.has_key('absent') ): self._period = config['absent']
        self.reset()

    # String representation for debug
    def __str__  ( self ): return self._name

    # Reset monitor (usually after resume)
    def reset  ( self ): self._activity = time.time()

    # Check if monitored resource is active
    def active ( self):
        ret      = False
        inactive = time.time() - self._activity
        if ( inactive < self._period ):
            debug('%s - inactive for %0.2f secs' % (str(self), inactive))
            ret = True
        return ret

    # Get preferred grace period
    def grace  ( self ): return self._grace

    # Start the monitor
    def start  ( self ): pass

    # Stop the monitor
    def stop   ( self ): pass

# ###########################################################################
# Editor directives
# ###########################################################################

# vim:sts=4:ts=4:sw=4:et



#    powernapd plugin - Monitors a UDP socket for data
#
#    Copyright (C) 2011 Canonical Ltd.
#
#    Authors: Dustin Kirkland <kirkland@canonical.com>
#             Andres Rodriguez <andreserl@canonical.com>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import threading, time, socket, os, re, struct, traceback
from logging import error, debug, info, warn

# Obtain MAC address from Monitor Interface
def get_mac_address(iface):
    file = "/sys/class/net/%s/address" % iface
    f = open(file, 'r')
    iface = f.read()
    f.close
    return iface.strip()

# Generate WoL data for local interface to compare with received packet (Partially taken from powerwake)
def get_local_wol_data(mac):
    nonhex = re.compile('[^0-9a-fA-F]')
    mac = nonhex.sub('', mac)
    if len(mac) != 12:
        error("Malformed mac address [%s]" % mac)
    data = ''.join(['FFFFFFFFFFFF', mac * 16])
    wol_data = ''
    for i in range(0, len(data), 2):
        wol_data = ''.join([wol_data, struct.pack('B', int(data[i: i + 2], 16))])
    return wol_data

# Obtain a list of available eth's, with its MAC address and WoL data.
def get_eths_mac_wol_info():
    ifaces = []
    prefix = re.compile("eth")
    dirs = os.listdir("/sys/class/net")
    for iface in dirs:
        if prefix.search(iface):
            # Obtain MAC address
            mac = get_mac_address(iface)
            # Obtain WoL data of eth
            data = get_local_wol_data(mac)
            ifaces.append({"iface":iface, "mac":mac, "wol":data})
    return ifaces

# Monitor plugin
#   listen for WoL data in a UDP socket. It compares if the data is specifically
#   for any of the interfaces
class WoLMonitor (threading.Thread):

    # Initialise
    def __init__ ( self, config ):
        threading.Thread.__init__(self)
        self._type = config['monitor']
        self._name = config['name']
        self._port = config['port']
        self._host = '' # Bind to all Interfaces
        self._running = False
        self._data_received = False
        self._absent_seconds = 0

    # Start thread
    def start ( self ):
      self._running = True
      threading.Thread.start(self)

    # Stop thread
    def stop ( self ): self._running = False

    # Open port and wait for data (any data will trigger the monitor)
    def run ( self ):

        isRunning = False
        #self._port = 7
        ifaces = get_eths_mac_wol_info()

        # Prepare the socket and bind port
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        try:
            s.bind((self._host, self._port))
            isRunning = True
        except:
            #error("Unable to bind port [%s]" % port)
            return

        #while isRunning:
        while self._running:
            try:
                #debug(logging.DEBUG, "    WoL monitor started at port [%s]" % port)
                recv_wol_msg, address = s.recvfrom(1024)
                #debug(logging.DEBUG, "    WoL packet received from %s" % address[0])
                for iface in ifaces:
                    if recv_wol_msg == iface["wol"]:
                        #debug(logging.DEBUG, "    WoL data matches local interface [%s]" % iface["iface"])
                        self._data_received = True
                        #isRunning = False
			# TODO: Should return signal to daemon and wake up???
                        #break
            except (KeyboardInterrupt, SystemExit):
                raise
            except:
                traceback.print_exc()

    def active(self):
        if self._data_received:
            self._data_received = False
            return True
        return False

# ###########################################################################
# Editor directives
# ###########################################################################

# vim:sts=4:ts=4:sw=4:et



#    powernapd plugin - Monitors /dev/input for user activity
#
#    Copyright (C) 2010, 2011 Canonical Ltd.
#
#    Authors: Dustin Kirkland <kirkland@canonical.com>
#             Adam Sutton <dev@adamsutton.me.uk>
#             Andres Rodriguez <andreserl@canonical.com>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os, re, threading
import select
from fcntl import fcntl, F_NOTIFY, DN_CREATE, DN_DELETE, DN_MULTISHOT
from logging import error, debug, info, warn

# Monitor plugin
#   Monitors devices in /dev/input for activity
class InputMonitor ( threading.Thread ):

    # Initialise
    def __init__ ( self, config ):
        threading.Thread.__init__(self)
        self._type = config['monitor']
        self._name = config['name']
        self._absent_seconds = 0
        self._input_received = False

        # If regex is in the way of by-id/regex, then path is changed to /dev/input/by-id
        if os.path.split(config['regex'])[0]:
            self._path = os.path.join("/dev/input", os.path.dirname(config['regex']))
            self._regex = re.compile(os.path.basename(config['regex']))
        elif config["regex"] == "kbd":
            self._path = "/dev/input/by-id"
            self._regex = re.compile(config['regex'])
        else:
            self._path = "/dev/input"
            self._regex = re.compile(config['regex'])

        # Register for directory events / setup input watches
        self._inputs = {}
        self._poll   = select.poll()
        self._update_inputs()
        self._dd     = os.open(self._path, 0)
        fcntl(self._dd, F_NOTIFY, DN_DELETE | DN_CREATE | DN_MULTISHOT)

    # Update the input events list
    def _update_inputs ( self ):
        events = {}
        event = None # Name of the event to poll, after finding it
        evpath = None # Absolute path of the event to poll

        # Search for the event in path:
        for str in os.listdir(self._path):
            match = self._regex.search(str)
            if match:
                event = str
                evpath = os.path.abspath(os.path.join(self._path, event))
                break

        # If event is different from None, then update!
        if event:
	    if os.path.exists(evpath):
		fp = open(evpath)
		events[evpath] = fp
		# Register the event to poll
		self._poll.register(fp.fileno(), select.POLLIN|select.POLLPRI)
		self._inputs = events

    # Start the thread
    def start ( self ):
        self._running = True
        threading.Thread.start(self)

    # Stop thread
    def stop ( self ): self._running = False

    # Monitor /dev/input
    def run ( self ):

        # Poll for events
        while self._running:
            res = self._poll.poll(1000)
            if ( res ):
                for fd, e in res:
                    if e & (select.POLLIN|select.POLLPRI):
                        os.read(fd, 32768) # Read what is there!
                        self._input_received = True

    def active(self):
        if self._input_received:
            self._input_received = False
            return True
        return False

# ###########################################################################
# Editor directives
# ###########################################################################

# vim:sts=4:ts=4:sw=4:et



#    powernapd plugin - Monitors system load
#
#    Copyright (C) 2011 Canonical Ltd.
#
#    Authors: Dustin Kirkland <kirkland@ubuntu.com>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os, multiprocessing
from logging import error, debug, info, warn

class LoadMonitor():

    # Initialise
    def __init__(self, config):
        self._type = config['monitor']
        self._name = config['name']
        self._threshold = config['threshold']
        self._absent_seconds = 0

    # Check system load
    def active(self):
        t = self._threshold
        if t == "n":
            t = multiprocessing.cpu_count()
        if os.getloadavg()[0] > float(t):
	    return True
	return False

    def start(self):
        pass

# ###########################################################################
# Editor directives
# ###########################################################################

# vim:sts=4:ts=4:sw=4:et



#!/usr/bin/env python
#
# Script to migrate a v1.7 configuration file to a v1.xx configuration
# file

import os, sys

# Parse command line
if len(sys.argv) != 4:
  print 'Usage :- %s old_config old_action new_config' % sys.argv[0]
old_cnf_path = sys.argv[1]
old_act_path = sys.argv[2]
new_cnf_path = sys.argv[3]

# Load OLD config
try:
  execfile(old_cnf_path)
except Exception, e:
  print 'ERROR: failed to load old configuration [e=%s]' % e
  sys.exit(1)

# Process new config file
lines = []
try:
  fp = open(new_cnf_path, 'r')
  for l in fp.readlines():
    l = l.strip()

    # Ignore blank, comment and section line
    if not l or l[0] in [ '#', '[' ]:
      lines.append(l)
   
    # Interval
    elif l.startswith('interval'):
      lines.append('interval = %0.2f' % INTERVAL_SECONDS)

    # Absent
    elif l.startswith('absent'):
      lines.append('absent = %0.2f' % ABSENT_SECONDS)
 
    # Grace
    elif l.startswith('grace'):
      lines.append('grace = %0.2f' % GRACE_SECONDS)

    # Debug
    elif l.startswith('debug'):
      lines.append('debug = %d' % DEBUG)
  
    # Action
    elif l.startswith('action_enter_sleep'):
      if os.access(old_act_path, os.X_OK):
        lines.append("action_enter_sleep = '%s'" % os.path.abspath(old_act_path))
      else: lines.append(l)

    # Monitors
    elif l.startswith('monitors'):
      monitors = "monitors = [ \"InputMonitor()\""
      for m in MONITORED_PROCESSES:
        monitors += ", \"ProcessMonitor({'regex':'%s'})\"" % m
      monitors += " ]"
      lines.append(monitors)

    # All else
    else: lines.append(l)
  fp.close()

  # Store data
  fp = open(new_cnf_path, 'w')
  for l in lines:
    fp.write(l + '\n')
  fp.close()
except Exception, e:
  print 'ERROR: failed to update configuration [e=%s]' % e
  sys.exit(1)

# Done
print 'Configuration successfully updated'



#!/usr/bin/python
#
#    powerwake.py - handles powerwaked config and initializes Monitors.
#
#    Copyright (C) 2011 Canonical Ltd.
#
#    Authors: Andres Rodriguez <andreserl@canonical.com>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import ConfigParser, sys, re, os
#from monitors import ARPMonitor 

class PowerWake:

    def __init__(self):
        self.PKG = "powerwake"
        self.CONFIG = "/etc/powernap/powerwaked.conf"
        self.ACTION = "/usr/bin/powerwake"
        self.INTERVAL_SECONDS = int(1)
        self.DEBUG = int(0)
        self.MONITORS = []
        # Load default config file (/etc/powernap/config)
        self.load_config_file()

    def load_config_file(self):
        cfg = ConfigParser.ConfigParser()
        cfg.read(self.CONFIG)

        try:
            # Load items in DEFAULT section
            defaults = cfg.items(self.PKG)
            for items in defaults:
                self.set_default_values(items[0], items[1])

            monitors_config = cfg.sections()
            for monitor in monitors_config:
                for items in cfg.items(monitor):
                    self.load_monitors_config(monitor, items)
        except:
            pass

    def set_default_values(self, var, value):
        if var == "interval_seconds":
            self.INTERVAL_SECONDS = eval(value)
        if var == "debug":
            self.DEBUG = eval(value)
        if var == "action":
            self.ACTION = eval(value)
        if var == "warn":
            if value == "y" or value == "yes":
                self.WARN = True

    def load_monitors_config(self, monitor, items):
        if monitor == "ARPMonitor" and (items[1] == "y" or items[1] == "yes"):
            self.MONITORS.append({"monitor":monitor, "cache":self.get_monitored_hosts(monitor.lower())})

    def get_monitors(self):
        from monitors import ARPMonitor 
        monitor = []
        for config in self.MONITORS:
            if config["monitor"] == "ARPMonitor":
                p = ARPMonitor.ARPMonitor(config)
            monitor.append(p)

        return monitor

    def get_monitored_hosts(self, monitor):
        host_to_mac = {}
        for file in ["/etc/powernap/powerwaked.%s.ethers" % monitor]:
            if os.path.exists(file):
                f = open(file, 'r')
                for i in f.readlines():
                    try:
                        (m, h) = i.split()
                        host_to_mac[h] = m
                    except:
                        pass
                f.close()
        return host_to_mac

    def set_monitored_hosts(self, host_to_mac, monitor):
        path = ["/etc/powernap/powerwaked.%s.ethers" % monitor]
        for file in path:
            if not os.path.exists(file):
                f = open(file, 'a')
                f.close()
        for file in ["/etc/powernap/powerwaked.%s.ethers" % monitor]:
            if os.access(file, os.W_OK):
                f = open(file, 'w')
                for h in host_to_mac:
                    if self.is_mac(host_to_mac[h]):
                        f.write("%s %s\n" % (host_to_mac[h], h))
                f.close()

    def get_mac_or_ip_from_arp(self, host):
        mac_or_ip = None
        for i in os.popen("/usr/sbin/arp -n"):
            m = i.split()[2]
            h = i.split()[0]
            if self.is_mac(host) and host == m and self.is_ip(h):
                mac_or_ip = h
                break
            if self.is_ip(host) and host == h and self.is_mac(m):
                mac_or_ip = m
                break
        #if not mac_or_ip:
        #    raise BaseException("Error")
        return mac_or_ip

    def is_ip(self, ip):
        r1 = re.compile('^[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+$')
        if r1.match(ip):
            return True
        else:
            return False

    def is_mac(self, mac):
        r1 = re.compile('^[0-9a-fA-F]{12}$')
        r2 = re.compile('^[0-9a-fA-F]{2}.[0-9a-fA-F]{2}.[0-9a-fA-F]{2}.[0-9a-fA-F]{2}.[0-9a-fA-F]{2}.[0-9a-fA-F]{2}$')
        if r1.match(mac) or r2.match(mac):
            return 1
        else:
            return 0

#### --------------------------------- from powerwake --------------------------------####

    # Source the cached, known arp entries
    def get_arp_cache(self):
        host_to_mac = {}
        #for file in ["/var/cache/%s/ethers" % self.PKG, "/etc/ethers", "%s/.cache/ethers" % HOME]:
        for file in ["/var/cache/%s/ethers" % self.PKG, "/etc/ethers"]:
            if os.path.exists(file):
                f = open(file, 'r')
                for i in f.readlines():
                    try:
                        (m, h) = i.split()
                        host_to_mac[h] = m
                    except:
                        pass
                f.close()
        return host_to_mac

    # Source the current, working arp table
    def get_arp_current(self, host_to_mac):
        # Load hostnames
        for i in os.popen("/usr/sbin/arp"):
            m = i.split()[2]
            h = i.split()[0]
            if is_mac(m):
                host_to_mac[h] = m
        # Load ip addresses
        for i in os.popen("/usr/sbin/arp -n"):
            m = i.split()[2]
            h = i.split()[0]
            if is_mac(m):
                host_to_mac[h] = m
        return host_to_mac

    def write_arp_cache(self, host_to_mac):
        if not os.access("%s/.cache/" % HOME, os.W_OK):
            return
        if not os.path.exists("%s/.cache/" % HOME):
            os.makedirs("%s/.cache/" % HOME)
        f = open("%s/.cache/ethers" % HOME, 'a')
        f.close()
        for file in ["/var/cache/%s/ethers" % PKG, "%s/.cache/ethers" % HOME]:
            if os.access(file, os.W_OK):
                f = open(file, 'w')
                for h in host_to_mac:
                    if self.is_mac(host_to_mac[h]):
                        f.write("%s %s\n" % (host_to_mac[h], h))
                f.close()



#    powernapd plugin - Monitors disk power state
#
#    Copyright (C) 2011 Jim Heck.
#
#    Authors: Jim Heck <pinball.rules@gmail.com>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
import re, commands
from logging import error, debug, info, warn

# Monitor plugin
#   looks for disks that are active/idle.  Useful for sleeping only when
#   specified disks are in standby

class DiskMonitor ():

    # Initialise
    def __init__(self, config):
        self._type = config['monitor']
        self._name = config['name']
        self._regex_state = re.compile(r"^\s+drive\s+state\s+is:\s+(\S+)")
        self._regex_not_found = re.compile(r"^.*No\s+such\s+file")
        self._absent_seconds = 0

    def start(self):
        pass

    def active(self):
        if self.is_disk_active():
            return True
        return False

    # Check for inactive drive by looking explicitly for drive state of
    # 'standby' or 'sleeping'.  Assume 'active/idle', except in case
    # where fuction returns 'No such file' error (e.g. unknown drive)
    def is_disk_active(self):
        hdparm = commands.getoutput("hdparm -C /dev/%s" % self._name).splitlines()
        is_active = True
        for line in hdparm:
            if self._regex_not_found.match(line):
                #warn("    Disk monitor: disk %s not found, ignoring" % self._name)
                return False
            if self._regex_state.match(line):
                state = self._regex_state.search(line).group(1)
                #debug("    Disk monitor: disk %s in state %s" % (self._name, state))
                if state == 'standby' or state == 'sleeping':
                    is_active = False
        if is_active:
            return True
        return False


# ###########################################################################
# Editor directives
# ###########################################################################

# vim:sts=4:ts=4:sw=4:et



#    powernapd plugin - Monitors process table for presence of process
#
#    Copyright (C) 2011 Canonical Ltd.
#
#    Authors: Dustin Kirkland <kirkland@canonical.com>
#             Andres Rodriguez <andreserl@canonical.com>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os, re, commands, time
from logging import error, debug, info, warn

# Check /dev/*, such that we don't powernap the system if someone
# is actively using a terminal device
def get_console_activity():
    ptmx = "/dev/ptmx"
    time = os.stat(ptmx).st_mtime
    irqs = get_interrupts()
    return time, irqs

# Obtain the interrupts at any given point in time
def get_interrupts():
    interrupts = 0
    f = open("/proc/interrupts", "r")
    for line in f.readlines():
        items = line.split()
        source = items.pop()
        if source == "i8042" or source == "keyboard" or source == "mouse":
            items.pop(0)
            items.pop()
            for i in items:
                interrupts += int(i)
    f.close()
    return interrupts

class ConsoleMonitor():

    # Initialise
    def __init__(self, config):
        self._type = config['monitor']
        self._name = config['name']
        self._absent_seconds = 0
        self._time, self._irqs = get_console_activity()

    # Check for PIDs
    def active(self):
        cur_time, cur_irqs = get_console_activity()
        if cur_time > self._time or cur_irqs > self._irqs:
                self._irqs = cur_irqs
                self._time = cur_time
		return True
	return False

    def start(self):
        pass




#    powernapd plugin - Monitors process table for process with IO activity
#
#    Copyright (C) 2010, 2011 Canonical Ltd.
#
#    Authors: Dustin Kirkland <kirkland@canonical.com>
#             Adam Sutton <dev@adamsutton.me.uk>
#             Andres Rodriguez <andreserl@canonical.com>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
import re
from logging import error, debug, info, warn, os
#from Monitor import Monitor

# Monitor plugin
#   looks for processes that have IO activity. Useful for some server
#   processes that are always present in the process list even when idle

# Looks for the process regex in /proc/<PID>/cmdline to obtain its PID(s).
# Optimal for processes that have a command line.
def find_pids_cmdline(regex):
    ret = []
    for d in os.listdir('/proc'):
        try:
            path = '/proc/%s/cmdline' % d
            if os.path.isfile(path):
                fp      = open(path)
                cmdline = fp.read()
                fp.close()
                if regex.search(cmdline):
                    ret.append(int(d))
        except:
            pass
    return ret

# Looks for the process regex in the "Name:" filed of /proc/<PID>/status
# to obtain its PID(s). Optimal for processes that do NOT have a command
# line. (i.e. NFS daemon processes.)
def find_pids_status(regex):
    ret = []
    for d in os.listdir('/proc'):
        try:
            path = '/proc/%s/status' % d
            if os.path.isfile(path):
                fp = open(path)
                status = fp.readline()
                fp.close()
                if regex.search(status.split()[1]):
                    ret.append(int(d))
        except:
            pass
    return ret

class IOMonitor ():

    # Initialise
    def __init__ ( self, config ):
        self._iocounts = {}
        self._type = config['monitor']
        self._name = config["name"]
        self._regex = re.compile(config['regex'])
        self._absent_seconds = 0

    def start(self):
        pass

    def active(self):
        if self.get_io_count():
            return True
        return False

    # Check for activity
    def get_io_count ( self ):

        # Get new PID list from processes with command line.
        pids = find_pids_cmdline(self._regex)
        # Processes with no command line result on an empty PID list.
        # if so, use alternate search method.
        if not pids:
            pids = find_pids_status(self._regex)

        # Get IO counts for all PIDs
        io_counts = {}
        for pid in pids:
            io_counts[pid] = {}
            try:
                fp = open('/proc/%d/io' % pid)
                for l in fp.readlines():
                    pts = l.split(':')
                    io_counts[pid][pts[0].strip()] = int(pts[1].strip())
                fp.close()
            except: pass # its possible the proc will die in here!

        ioactivity = False
        for pid in pids:
            # New process (assume activity)
            if pid not in self._iocounts:
                debug('    %s - adding new PID %d to list' % (self, pid))
            # Existing: check for change
            else:
                if (self._iocounts[pid]["write_bytes"] != io_counts[pid]["write_bytes"]) or \
                   (self._iocounts[pid]["read_bytes"] != io_counts[pid]["read_bytes"]):
                    ioactivity = True
                    self._iocounts[pid] = io_counts[pid]
                    break
            # Update count
            self._iocounts[pid] = io_counts[pid]

        if ioactivity:
            return True

        return False

# ###########################################################################
# Editor directives
# ###########################################################################

# vim:sts=4:ts=4:sw=4:et



#!/usr/bin/python
#
#    powernap.py - handles powernap's config and initializes Monitors.
#
#    Copyright (C) 2010 Canonical Ltd.
#
#    Authors: Andres Rodriguez <andreserl@ubuntu.com>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import ConfigParser, sys, re, os
from monitors import ProcessMonitor, LoadMonitor, InputMonitor, TCPMonitor, UDPMonitor, IOMonitor, WoLMonitor, ConsoleMonitor, DiskMonitor


class PowerNap:

    def __init__(self):
        self.PKG = "powernap"
        self.CONFIG = "/etc/powernap/config"
        self.ACTION = "/usr/sbin/powernap"
        self.RECOVER_ACTION = "/usr/sbin/pm-powersave false"
        self.ABSENT_SECONDS = sys.maxint
        self.STAGE2_ABSENT_SECONDS = sys.maxint
        self.INTERVAL_SECONDS = int(1)
        self.GRACE_SECONDS = int(60)
        self.DEBUG = int(0)
        self.ACTION_METHOD = 0
        self.STAGE2_ACTION_METHOD = 4
        self.MONITORS = []
        self.WARN = False
        self.WATCH_CONFIG = False
        self.KERN_MODULES = {}
        self.SERVICES = {}
        self.stage2_action_enabled = False
        # Load default config file (/etc/powernap/config)
        self.load_config_file()

    def load_config_file(self):
        stage2_section = "%s-stage2" % self.PKG
        cfg = ConfigParser.ConfigParser()
        cfg.read(self.CONFIG)

        try:
            # Load items in DEFAULT section
            defaults = cfg.items(self.PKG)
            for items in defaults:
                self.set_default_values(items[0], items[1])

            stage2 = cfg.items(stage2_section)
            for items in stage2:
                self.set_stage2_values(items[0], items[1])

            # Load items on each monitor
            monitors_config = cfg.sections()
            for monitor in monitors_config:
                if monitor not in [self.PKG, stage2_section]:
                    for items in cfg.items(monitor):
                        self.load_monitors_config(monitor, items)
        except:
            pass

        # Load extra config files (/etc/powernap/config.d/*)
        configd = "/etc/%s/config.d" % self.PKG
        if os.path.exists(configd):
            for config in os.listdir(configd):
                self.load_configd_files("%s/%s" % (configd, config))

    def load_configd_files(self, config_file):
        cfg = ConfigParser.ConfigParser()
        cfg.read(config_file)

        try:
            monitors_config = cfg.sections()
            for monitor in monitors_config:
                for items in cfg.items(monitor):
                    for i in range(len(self.MONITORS)):
                        if self.MONITORS[i]['monitor'] == monitor and self.MONITORS[i]['name'] == items[0]:
                            self.MONITORS.pop(i)
                            break
                    self.load_monitors_config(monitor, items)
        except:
            pass

    def set_default_values(self, var, value):
        if var == "absent_seconds":
            self.ABSENT_SECONDS = eval(value)
        if var == "interval_seconds":
            self.INTERVAL_SECONDS = eval(value)
        if var == "grace_seconds":
            self.GRACE_SECONDS = eval(value)
        if var == "debug":
            self.DEBUG = eval(value)
        if var == "action":
            self.ACTION = eval(value)
        if var == "action_method":
            self.ACTION_METHOD = eval(value)
        if var == "warn":
            if value == "y" or value == "yes":
                self.WARN = True
            else:
                self.WARN = False
        if var == "watch_config":
            if value == "y" or value == "yes":
                self.WATCH_CONFIG = True
            else:
                self.WATCH_CONFIG = False
        if var == "kern_modules":
            self.KERN_MODULES = value.split()
        if var == "services":
            self.SERVICES = value.split()

    def set_stage2_values(self, var, value):
        if var == "stage2_action_method":
            self.STAGE2_ACTION_METHOD = eval(value)
        if var == "stage2_absent_seconds":
            self.STAGE2_ABSENT_SECONDS = eval(value)
            if self.STAGE2_ABSENT_SECONDS > 0:
                self.stage2_action_enabled = True

    def usb_input_available(self, event):
        regex = re.compile(event)
        path = "/dev/input/by-id"
        exists = False

        for str in os.listdir(path):
            match = regex.search(str)
            if match:
                exists = True
                break
        return exists

    def load_monitors_config(self, monitor, items):
        if monitor == "ProcessMonitor" or monitor == "IOMonitor":
            self.MONITORS.append({"monitor":monitor, "name":items[0], "regex":eval(items[1]), "absent":self.ABSENT_SECONDS})
        if monitor == "InputMonitor" and (items[1] == "y" or items[1] == "yes"):
            if items[0] == "mouse" and self.usb_input_available("mouse"):
                self.MONITORS.append({"monitor":monitor, "name":items[0], "regex":"mice"})
            elif items[0] == "keyboard" and self.usb_input_available("kbd"):
                self.MONITORS.append({"monitor":monitor, "name":items[0], "regex":"kbd"})
            #else:
            #    self.MONITORS.append({"monitor":monitor, "name":items[0], "regex":items[1]})
        if monitor == "ConsoleMonitor" and (items[1] == "y" or items[1] == "yes"):
            self.MONITORS.append({"monitor":monitor, "name":items[0]})
        if monitor == "LoadMonitor":
            self.MONITORS.append({"monitor":monitor, "name":items[0], "threshold":items[1]})
        if monitor == "TCPMonitor":
            self.MONITORS.append({"monitor":monitor, "name":items[0], "port":items[1], "absent":self.ABSENT_SECONDS})
        if monitor == "UDPMonitor":
            # If ACTION_METHOD is 0 (PowerSave) and port is 7 or 9, do *NOT* create a monitor
            # This will cause that the WoL monitor to not be able to bind the port or viceversa.
            # TODO: Display a message that port is not being binded!!
            if self.ACTION_METHOD == 0 and (items[1] != 7 or items[1] != 9):
                self.MONITORS.append({"monitor":monitor, "name":items[0], "port":eval(items[1]), "absent":self.ABSENT_SECONDS})
        if monitor == "WoLMonitor":
            self.MONITORS.append({"monitor":monitor, "name":items[0], "port":eval(items[1]), "absent":self.ABSENT_SECONDS})
        if monitor == "DiskMonitor" and (items[1] == "y" or items[1] == "yes"):
            self.MONITORS.append({"monitor":monitor, "name":items[0], "absent":self.ABSENT_SECONDS})

    def get_monitors(self):
        monitor = []
        for config in self.MONITORS:
            if config["monitor"] == "ProcessMonitor":
                p = ProcessMonitor.ProcessMonitor(config)
            if config["monitor"] == "LoadMonitor":
                p = LoadMonitor.LoadMonitor(config)
            if config["monitor"] == "UDPMonitor":
                p = UDPMonitor.UDPMonitor(config)
            if config["monitor"] == "WoLMonitor":
                p = WoLMonitor.WoLMonitor(config)
            if config["monitor"] == "InputMonitor":
                p = InputMonitor.InputMonitor(config)
            if config["monitor"] == "ConsoleMonitor":
                p = ConsoleMonitor.ConsoleMonitor(config)
            if config["monitor"] == "IOMonitor":
                p = IOMonitor.IOMonitor(config)
            if config["monitor"] == "TCPMonitor":
                p = TCPMonitor.TCPMonitor(config)
            if config["monitor"] == "DiskMonitor":
                p = DiskMonitor.DiskMonitor(config)
            monitor.append(p)

        return monitor



#
#    powernapd plugin - ARP Monitor for auto-wakeup of client machines
#
#    Copyright (C) 2011 Canonical Ltd.
#
#    Authors: Andres Rodriguez <andreserl@canonical.com>
#             Jim Heck <pinball.rules@gmail.com>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os, threading
from logging import error, debug, info, warn
from scapy.all import *
#from scapy.all import sniff, ARP, Ether, IP, UDP, Raw, sendp

# Converts the MAC to some weird format
def mac_to_string(mac):
    return str(Ether(dst=mac))[:6]

# Create a WoL magic packet
def make_magic(sleeper_mac_str):
    return Ether(dst='ff:ff:ff:ff:ff:ff')/IP(dst='255.255.255.255')/UDP(dport=7)/Raw(load='\xff\xff\xff\xff\xff\xff' + sleeper_mac_str * 16)

class ARPMonitor(threading.Thread):

    # Initialise
    def __init__(self, config):
        threading.Thread.__init__(self)
        self._type = config['monitor']
        self._name = "ARP Monitor"
        self._arp_cache = config['cache']
        self._absent_seconds = 0
        self._running = False

    # Check for PIDs
    def active(self):
	return False 

    # Start thread
    def start(self):
        self._running = True
        threading.Thread.start(self)

    # Stop thread
    def stop(self):
        self._running = False

    def run(self):
        while self._running:
            sniff(prn=self.arp_wake_sleeper_callback, filter="arp", store=0)

    def arp_wake_sleeper_callback(self, pkt):
        # evaluates if received ARP packet with dest ip (pkt[ARP].pdst) is in cache
        if ARP in pkt and pkt[ARP].op == 1 and pkt[ARP].pdst in self._arp_cache:
            # If found in arp_cache, then try to wake up by sending a WoL.
            magic_pkt = make_magic(mac_to_string(self._arp_cache[pkt[ARP].pdst]))
            sendp(magic_pkt)
